/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#ifndef TCP_IPV6_SEGMENT_H
#define TCP_IPV6_SEGMENT_H

#include "ip/ipv6header.h"
#include "tcp/tcpheader.h"
#include "tcp/tcpsegment.h"

//-----------------------------------------------------------------------------
// TcpIpv6Segment
//-----------------------------------------------------------------------------

/**
 * @brief This class represents the TCP segment over IP version 6.
 *
 * The TCP Segment over IPv6 Packet format is:
 *
 * @code
 * 0      3 4            11 12   15 16          23 24            31
 * +-------+---------------+--------------------------------------+      ---
 * |       |               |                                      |       ^
 * |version| traffic class |              flow label              |       |
 * |  (6)  |               |                                      |       |
 * +-------+---------------+-------+--------------+---------------+       |
 * |                               |              |               |       |
 * |        payload length         | next header  |   hop limit   |       |
 * |                               |              |               |       |
 * +---------------+---------------+--------------+---------------+       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                      source IPv6 address                     | IPv6 Header
 * |                                                              |   40 bytes
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                   destination IPv6 address                   |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       |
 * |                                                              |       v
 * +-------------------------------+------------------------------+      ---
 * |                               |                              |       ^
 * |         source port           |     destination port         |       |
 * |                               |                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |       |
 * |                        sequence number                       |       |
 * |                                                              |       |
 * +--------------------------------------------------------------+       |
 * |                                                              |  TCP Header
 * |              acknowledgment number (if ACK set)              |   20 bytes
 * |                                                              |       |
 * +-------+-------+-+-+-+-+-+-+-+-+------------------------------+       |
 * | data  | reser-|C|E|U|A|P|R|S|F|                              |       |
 * |offset | ved   |W|C|R|C|S|S|Y|I|        window size           |       |
 * |       |       |R|E|G|K|H|T|N|N|                              |       |
 * +---------------+-+-+-+-+-+-+-+-+------------------------------+       |
 * |                               |                              |       |
 * |          checksum             | urgent pointer (if URG set)  |       |
 * |                               |                              |       v
 * +-----------------------------------------------+--------------+      ---
 * @endcode
 */
class TcpIpv6Segment : public TcpSegment
{
public:
    TcpIpv6Segment();
    TcpIpv6Segment( const TcpHeader &tcp_header );
    virtual ~TcpIpv6Segment();

    Ipv6Header get_ip_header() const;
    TcpHeader get_tcp_header() const;

    virtual bool match_rst_reply(
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_rst_reply(
            const boost::posix_time::ptime &time_segment_sent
    ) const;

    virtual bool match_syn_ack_reply(
            const boost::asio::ip::address &source_address
    ) const;

    virtual void print_syn_ack_reply(
            const boost::posix_time::ptime &time_packet_sent
    ) const;

    virtual bool read( std::istream &is );
    virtual bool write( std::ostream &os ) const;

    friend std::istream& operator>>(
            std::istream &is,
            TcpIpv6Segment &segment
    );
    friend std::ostream& operator<<(
            std::ostream &os,
            const TcpIpv6Segment &segment
    );

private:
    /// The IP header.
    Ipv6Header IpHeader;
    /// The TCP header.
    TcpHeader TcpHdr;

};

#endif // TCP_IPV6_SEGMENT_H
